#ifdef BUFFALO_GRUB
/* URL http://forum.osdev.org/viewtopic.php?t=16990 */
//
// here is the slighlty complicated ACPI poweroff code
//

#include <grub/dl.h>
#include <grub/types.h>
#include <grub/mm.h>
#include <grub/acpi.h>
#include <grub/misc.h>
#include <grub/i386/io.h>
#include <grub/time.h>
#include <grub/err.h>
#include <grub/extcmd.h>
#include <grub/i18n.h>


static const struct grub_arg_option options[] =
  {
    {0, 0, 0, 0, 0, 0}
  };


grub_uint32_t SMI_CMD;
grub_uint8_t ACPI_ENABLE;
grub_uint8_t ACPI_DISABLE;
grub_uint32_t PM1a_CNT;
grub_uint32_t PM1b_CNT;
grub_uint16_t SLP_TYPa;
grub_uint16_t SLP_TYPb;
grub_uint16_t SLP_EN;
grub_uint16_t SCI_EN;
grub_uint8_t PM1_CNT_LEN;


struct RSDPtr
{
   grub_uint8_t Signature[8];
   grub_uint8_t CheckSum;
   grub_uint8_t OemID[6];
   grub_uint8_t Revision;
   grub_uint32_t *RsdtAddress;
};



struct FACP
{
   grub_uint8_t Signature[4];
   grub_uint32_t Length;
   grub_uint8_t unneded1[40 - 8];
   grub_uint32_t DSDT;
   grub_uint8_t unneded2[48 - 44];
   grub_uint32_t SMI_CMD;
   grub_uint8_t ACPI_ENABLE;
   grub_uint8_t ACPI_DISABLE;
   grub_uint8_t unneded3[64 - 54];
   grub_uint32_t PM1a_CNT_BLK;
   grub_uint32_t PM1b_CNT_BLK;
   grub_uint8_t unneded4[89 - 72];
   grub_uint8_t PM1_CNT_LEN;
} __attribute__ ((packed));



// check if the given address has a valid header
static
unsigned int *acpiCheckRSDPtr(unsigned int *ptr)
{
   char *sig = "RSD PTR ";
   struct RSDPtr *rsdp = (struct RSDPtr *) ptr;
   grub_uint8_t *bptr;
   grub_uint8_t check = 0;
   unsigned int i;

   if (grub_memcmp(sig, rsdp, 8) == 0)
   {
      // check checksum rsdpd
      bptr = (grub_uint8_t *) ptr;
      for (i=0; i<sizeof(struct RSDPtr); i++)
      {
         check += *bptr;
         bptr++;
      }

      // found valid rsdpd   
      if (check == 0) {
         /*
          if (desc->Revision == 0)
            grub_dprintf("acpi", "acpi 1");
         else
            grub_dprintf("acpi", "acpi 2");
         */
         return (unsigned int *) rsdp->RsdtAddress;
      }
   }

   return NULL;
}



// finds the acpi header and returns the address of the rsdt
static
unsigned int *acpiGetRSDPtr(void)
{
   unsigned int *addr;
   unsigned int *rsdp;

   // search below the 1mb mark for RSDP signature
   for (addr = (unsigned int *) 0x000E0000; (int) addr<0x00100000; addr += 0x10/sizeof(addr))
   {
      rsdp = acpiCheckRSDPtr(addr);
      if (rsdp != NULL)
         return rsdp;
   }


   // at address 0x40:0x0E is the RM segment of the ebda
   int ebda = *((short *) 0x40E);   // get pointer
   ebda = ebda*0x10 &0x000FFFFF;   // transform segment into linear address

   // search Extended BIOS Data Area for the Root System Description Pointer signature
   for (addr = (unsigned int *) ebda; (int) addr<ebda+1024; addr+= 0x10/sizeof(addr))
   {
      rsdp = acpiCheckRSDPtr(addr);
      if (rsdp != NULL)
         return rsdp;
   }

   return NULL;
}



// checks for a given header and validates checksum
static
int acpiCheckHeader(unsigned int *ptr, char *sig)
{
   if (grub_memcmp(ptr, sig, 4) == 0)
   {
      char *checkPtr = (char *) ptr;
      int len = *(ptr + 1);
      char check = 0;
      while (0<len--)
      {
         check += *checkPtr;
         checkPtr++;
      }
      if (check == 0)
         return 0;
   }
   return -1;
}


static
int acpiEnable(void)
{
    /* check if acpi is enabled */
    if ( (grub_inw((unsigned short int) PM1a_CNT) &SCI_EN) == 0 )
    {
      /* check if acpi can be enabled */
      if (SMI_CMD != 0 && ACPI_ENABLE != 0)
      {
         grub_outb(ACPI_ENABLE, (unsigned short int) SMI_CMD); /* send acpi enable command */
         /* give 3 seconds time to enable acpi */
         int i;
         for (i=0; i<300; i++ )
         {
            if ( (grub_inw((unsigned short int) PM1a_CNT) &SCI_EN) == 1 )
               break;
            grub_sleep(10);
         }
         if (PM1b_CNT != 0)
            for (; i<300; i++ )
            {
               if ( (grub_inw((unsigned short int) PM1b_CNT) &SCI_EN) == 1 )
                  break;
               grub_sleep(10);
            }
         if (i<300) {
            grub_dprintf("acpi", "enabled acpi.\n");
            return 0;
         } else {
            grub_dprintf("acpi", "couldn't enable acpi.\n");
            return -1;
         }
      } else {
         grub_dprintf("acpi", "no known way to enable acpi.\n");
         return -1;
      }
   } else {
      /* grub_dprintf("acpi", "acpi was already enabled.\n"); */
      return 0;
   }
}



//
// bytecode of the \_S5 object
// -----------------------------------------
//        | (optional) |    |    |    |   
// NameOP | \          | _  | S  | 5  | _
// 08     | 5A         | 5F | 53 | 35 | 5F
//
// -----------------------------------------------------------------------------------------------------------
//           |           |              | ( SLP_TYPa   ) | ( SLP_TYPb   ) | ( Reserved   ) | (Reserved    )
// PackageOP | PkgLength | NumElements  | byteprefix Num | byteprefix Num | byteprefix Num | byteprefix Num
// 12        | 0A        | 04           | 0A         05  | 0A          05 | 0A         05  | 0A         05
//
//----this-structure-was-also-seen----------------------
// PackageOP | PkgLength | NumElements |
// 12        | 06        | 04          | 00 00 00 00
//
// (Pkglength bit 6-7 encode additional PkgLength bytes [shouldn't be the case here])
//
static
int initAcpi(void)
{
   unsigned int *ptr = acpiGetRSDPtr();

   // check if address is correct  ( if acpi is available on this pc )
   if (ptr != NULL && acpiCheckHeader(ptr, "RSDT") == 0)
   {
      // the RSDT contains an unknown number of pointers to acpi tables
      int entrys = *(ptr + 1);
      entrys = (entrys-36) /4;
      ptr += 36/4;   // skip header information

      while (0<entrys--)
      {
         // check if the desired table is reached
         if (acpiCheckHeader((unsigned int *) *ptr, "FACP") == 0)
         {
            entrys = -2;
            struct FACP *facp = (struct FACP *) *ptr;
            if (acpiCheckHeader((unsigned int *) facp->DSDT, "DSDT") == 0)
            {
               // search the \_S5 package in the DSDT
               char *S5Addr = (char *) facp->DSDT +36; // skip header
               int dsdtLength = *((int *)facp->DSDT+1) -36;
               while (0 < dsdtLength--)
               {
                  if ( grub_memcmp(S5Addr, "_S5_", 4) == 0)
                     break;
                  S5Addr++;
               }
               // check if \_S5 was found
               if (dsdtLength > 0)
               {
                  // check for valid AML structure
                  if ( ( *(S5Addr-1) == 0x08 || ( *(S5Addr-2) == 0x08 && *(S5Addr-1) == '\\') ) && *(S5Addr+4) == 0x12 )
                  {
                     S5Addr += 5;
                     S5Addr += ((*S5Addr &0xC0)>>6) +2;   // calculate PkgLength size

                     if (*S5Addr == 0x0A)
                        S5Addr++;   // skip byteprefix
                     SLP_TYPa = *(S5Addr)<<10;
                     S5Addr++;

                     if (*S5Addr == 0x0A)
                        S5Addr++;   // skip byteprefix
                     SLP_TYPb = *(S5Addr)<<10;

                     SMI_CMD = facp->SMI_CMD;

                     ACPI_ENABLE = facp->ACPI_ENABLE;
                     ACPI_DISABLE = facp->ACPI_DISABLE;

                     PM1a_CNT = facp->PM1a_CNT_BLK;
                     PM1b_CNT = facp->PM1b_CNT_BLK;
                     
                     PM1_CNT_LEN = facp->PM1_CNT_LEN;

                     SLP_EN = 1<<13;
                     SCI_EN = 1;

                     return 0;
                  } else {
                     grub_dprintf("acpi", "\\_S5 parse error.\n");
                  }
               } else {
                  grub_dprintf("acpi", "\\_S5 not present.\n");
               }
            } else {
               grub_dprintf("acpi", "DSDT invalid.\n");
            }
         }
         ptr++;
      }
      grub_dprintf("acpi", "no valid FACP present.\n");
   } else {
      grub_dprintf("acpi", "no acpi.\n");
   }

   return -1;
}



static
void acpiPowerOff(void)
{
   /* SCI_EN is set to 1 if acpi shutdown is possible */
   if (SCI_EN == 0)
      return;

   acpiEnable();

   /* send the shutdown command */
   grub_outw(SLP_TYPa | SLP_EN, PM1a_CNT);
   if ( PM1b_CNT != 0 )
       grub_outw(SLP_TYPb | SLP_EN, PM1b_CNT);

   grub_printf("acpi poweroff failed.\n");
}


static grub_err_t
grub_cmd_acpi_poweroff (grub_extcmd_t cmd,
	      int argc, char **args)
{
    cmd = cmd;
    argc = argc;
    args = args;
     
    initAcpi();
    acpiPowerOff();

    return 0;
}

static grub_extcmd_t cmd;


GRUB_MOD_INIT(acpi_poweroff)
{
    cmd = grub_register_extcmd ("acpi_poweroff", 
                                grub_cmd_acpi_poweroff, GRUB_COMMAND_FLAG_BOTH,
                                N_(""), N_("ACPI soft-off (enter S5 state)"), 
                                options);
}

GRUB_MOD_FINI(acpi_poweroff)
{
  grub_unregister_extcmd (cmd);
}
#endif /* BUFFALO_GRUB */
